package main

import (
	"encoding/json"
	"fmt"
	"math"
	"net/http"
	"net/url"
	"time"
)

var tiIconNames = map[int]string{
	1000: "clear",
	1100: "mostlysunny",
	1101: "partlycloudy",
	1102: "mostlycloudy",
	1001: "cloudy",
	2000: "fog",
	2100: "hazy",
	4000: "rain",
	4200: "rain",
	4001: "rain",
	4201: "rain",
	8000: "tstorms",
	5001: "flurries",
	5100: "snow",
	5000: "snow",
	5101: "snow",
	7102: "sleet",
	7000: "sleet",
	7101: "sleet",
	6000: "sleet",
	6200: "sleet",
	6001: "sleet",
	6201: "sleet",
}

var tiDescriptions = map[int]string{
	1000: "Clear",
	1100: "Mostly sunny",
	1101: "Partly cloudy",
	1102: "Mostly cloudy",
	1001: "Cloudy",
	2000: "Fog",
	2100: "Light fog",
	4000: "Drizzle",
	4200: "Light rain",
	4001: "Rain",
	4201: "Heavy rain",
	8000: "Thunderstorms",
	5001: "Flurries",
	5100: "Light snow",
	5000: "Snow",
	5101: "Heavy snow",
	7102: "Light sleet",
	7000: "Sleet",
	7101: "Heavy sleet",
	6000: "Freezing drizzle",
	6200: "Light freezing rain",
	6001: "Freezing rain",
	6201: "Heavy freezing rain",
}

const tiAPI = "https://api.tomorrow.io/v4"

const tiUnits = "metric"

// TomorrowIo is a weather service handle
type TomorrowIo struct {
	apiKey string
}

type tiFloatValue struct {
	Value float64 `json:"value"`
	Units string  `json:"units"`
}

type tiIntValue struct {
	Value int    `json:"value"`
	Units string `json:"units"`
}

type tiStringValue struct {
	Value string `json:"value"`
}

type tiCurrent struct {
	Data struct {
		Time   string `json:"time"`
		Values struct {
			CloudBase                float64 `json:"cloudBase"`
			CloudCeiling             float64 `json:"cloudCeiling"`
			CloudCover               float64 `json:"cloudCover"`
			DewPoint                 float64 `json:"dewPoint"`
			FreezingRainIntensity    float64 `json:"freezingRainIntensity"`
			Humidity                 float64 `json:"humidity"`
			PrecipitationProbability float64 `json:"precipitationProbability"`
			PressureSurfaceLevel     float64 `json:"pressureSurfaceLevel"`
			RainIntensity            float64 `json:"rainIntensity"`
			SleetIntensity           float64 `json:"sleetIntensity"`
			SnowIntensity            float64 `json:"snowIntensity"`
			Temperature              float64 `json:"temperature"`
			TemperatureApparent      float64 `json:"temperatureApparent"`
			UvHealthConcern          int     `json:"uvHealthConcern"`
			UvIndex                  int     `json:"uvIndex"`
			Visibility               float64 `json:"visibility"`
			WeatherCode              int     `json:"weatherCode"`
			WindDirection            float64 `json:"windDirection"`
			WindGust                 float64 `json:"windGust"`
			WindSpeed                float64 `json:"windSpeed"`
		} `json:"values"`
	} `json:"data"`
	Location struct {
		Lat  float64 `json:"lat"`
		Lon  float64 `json:"lon"`
		Name string  `json:"name"`
		Type string  `json:"type"`
	} `json:"location"`
}

type tiForecast struct {
	Timelines struct {
		Hourly []struct {
			Time   string `json:"time"`
			Values struct {
				CloudBase                float64 `json:"cloudBase"`
				CloudCeiling             float64 `json:"cloudCeiling"`
				CloudCover               float64 `json:"cloudCover"`
				DewPoint                 float64 `json:"dewPoint"`
				Evapotranspiration       float64 `json:"evapotranspiration"`
				FreezingRainIntensity    float64 `json:"freezingRainIntensity"`
				Humidity                 float64 `json:"humidity"`
				IceAccumulation          float64 `json:"iceAccumulation"`
				IceAccumulationLwe       float64 `json:"iceAccumulationLwe"`
				PrecipitationProbability float64 `json:"precipitationProbability"`
				PressureSurfaceLevel     float64 `json:"pressureSurfaceLevel"`
				RainAccumulation         float64 `json:"rainAccumulation"`
				RainAccumulationLwe      float64 `json:"rainAccumulationLwe"`
				RainIntensity            float64 `json:"rainIntensity"`
				SleetAccumulation        float64 `json:"sleetAccumulation"`
				SleetAccumulationLwe     float64 `json:"sleetAccumulationLwe"`
				SleetIntensity           float64 `json:"sleetIntensity"`
				SnowAccumulation         float64 `json:"snowAccumulation"`
				SnowAccumulationLwe      float64 `json:"snowAccumulationLwe"`
				SnowIntensity            float64 `json:"snowIntensity"`
				Temperature              float64 `json:"temperature"`
				TemperatureApparent      float64 `json:"temperatureApparent"`
				UvHealthConcern          int     `json:"uvHealthConcern"`
				UvIndex                  int     `json:"uvIndex"`
				Visibility               float64 `json:"visibility"`
				WeatherCode              int     `json:"weatherCode"`
				WindDirection            float64 `json:"windDirection"`
				WindGust                 float64 `json:"windGust"`
				WindSpeed                float64 `json:"windSpeed"`
			} `json:"values"`
		} `json:"hourly"`
		Daily []struct {
			Time   string `json:"time"`
			Values struct {
				CloudBaseAvg                float64 `json:"cloudBaseAvg"`
				CloudBaseMax                float64 `json:"cloudBaseMax"`
				CloudBaseMin                float64 `json:"cloudBaseMin"`
				CloudCeilingAvg             float64 `json:"cloudCeilingAvg"`
				CloudCeilingMax             float64 `json:"cloudCeilingMax"`
				CloudCeilingMin             float64 `json:"cloudCeilingMin"`
				CloudCoverAvg               float64 `json:"cloudCoverAvg"`
				CloudCoverMax               float64 `json:"cloudCoverMax"`
				CloudCoverMin               float64 `json:"cloudCoverMin"`
				DewPointAvg                 float64 `json:"dewPointAvg"`
				DewPointMax                 float64 `json:"dewPointMax"`
				DewPointMin                 float64 `json:"dewPointMin"`
				EvapotranspirationAvg       float64 `json:"evapotranspirationAvg"`
				EvapotranspirationMax       float64 `json:"evapotranspirationMax"`
				EvapotranspirationMin       float64 `json:"evapotranspirationMin"`
				EvapotranspirationSum       float64 `json:"evapotranspirationSum"`
				FreezingRainIntensityAvg    float64 `json:"freezingRainIntensityAvg"`
				FreezingRainIntensityMax    float64 `json:"freezingRainIntensityMax"`
				FreezingRainIntensityMin    float64 `json:"freezingRainIntensityMin"`
				HumidityAvg                 float64 `json:"humidityAvg"`
				HumidityMax                 float64 `json:"humidityMax"`
				HumidityMin                 float64 `json:"humidityMin"`
				IceAccumulationAvg          float64 `json:"iceAccumulationAvg"`
				IceAccumulationLweAvg       float64 `json:"iceAccumulationLweAvg"`
				IceAccumulationLweMax       float64 `json:"iceAccumulationLweMax"`
				IceAccumulationLweMin       float64 `json:"iceAccumulationLweMin"`
				IceAccumulationMax          float64 `json:"iceAccumulationMax"`
				IceAccumulationMin          float64 `json:"iceAccumulationMin"`
				IceAccumulationSum          float64 `json:"iceAccumulationSum"`
				MoonriseTime                string  `json:"moonriseTime"`
				MoonsetTime                 string  `json:"moonsetTime"`
				PrecipitationProbabilityAvg float64 `json:"precipitationProbabilityAvg"`
				PrecipitationProbabilityMax float64 `json:"precipitationProbabilityMax"`
				PrecipitationProbabilityMin float64 `json:"precipitationProbabilityMin"`
				PressureSurfaceLevelAvg     float64 `json:"pressureSurfaceLevelAvg"`
				PressureSurfaceLevelMax     float64 `json:"pressureSurfaceLevelMax"`
				PressureSurfaceLevelMin     float64 `json:"pressureSurfaceLevelMin"`
				RainAccumulationAvg         float64 `json:"rainAccumulationAvg"`
				RainAccumulationLweAvg      float64 `json:"rainAccumulationLweAvg"`
				RainAccumulationLweMax      float64 `json:"rainAccumulationLweMax"`
				RainAccumulationLweMin      float64 `json:"rainAccumulationLweMin"`
				RainAccumulationMax         float64 `json:"rainAccumulationMax"`
				RainAccumulationMin         float64 `json:"rainAccumulationMin"`
				RainAccumulationSum         float64 `json:"rainAccumulationSum"`
				RainIntensityAvg            float64 `json:"rainIntensityAvg"`
				RainIntensityMax            float64 `json:"rainIntensityMax"`
				RainIntensityMin            float64 `json:"rainIntensityMin"`
				SleetAccumulationAvg        float64 `json:"sleetAccumulationAvg"`
				SleetAccumulationLweAvg     float64 `json:"sleetAccumulationLweAvg"`
				SleetAccumulationLweMax     float64 `json:"sleetAccumulationLweMax"`
				SleetAccumulationLweMin     float64 `json:"sleetAccumulationLweMin"`
				SleetAccumulationMax        float64 `json:"sleetAccumulationMax"`
				SleetAccumulationMin        float64 `json:"sleetAccumulationMin"`
				SleetIntensityAvg           float64 `json:"sleetIntensityAvg"`
				SleetIntensityMax           float64 `json:"sleetIntensityMax"`
				SleetIntensityMin           float64 `json:"sleetIntensityMin"`
				SnowAccumulationAvg         float64 `json:"snowAccumulationAvg"`
				SnowAccumulationLweAvg      float64 `json:"snowAccumulationLweAvg"`
				SnowAccumulationLweMax      float64 `json:"snowAccumulationLweMax"`
				SnowAccumulationLweMin      float64 `json:"snowAccumulationLweMin"`
				SnowAccumulationMax         float64 `json:"snowAccumulationMax"`
				SnowAccumulationMin         float64 `json:"snowAccumulationMin"`
				SnowAccumulationSum         float64 `json:"snowAccumulationSum"`
				SnowIntensityAvg            float64 `json:"snowIntensityAvg"`
				SnowIntensityMax            float64 `json:"snowIntensityMax"`
				SnowIntensityMin            float64 `json:"snowIntensityMin"`
				SunriseTime                 string  `json:"sunriseTime"`
				SunsetTime                  string  `json:"sunsetTime"`
				TemperatureApparentAvg      float64 `json:"temperatureApparentAvg"`
				TemperatureApparentMax      float64 `json:"temperatureApparentMax"`
				TemperatureApparentMin      float64 `json:"temperatureApparentMin"`
				TemperatureAvg              float64 `json:"temperatureAvg"`
				TemperatureMax              float64 `json:"temperatureMax"`
				TemperatureMin              float64 `json:"temperatureMin"`
				UvHealthConcernAvg          int     `json:"uvHealthConcernAvg"`
				UvHealthConcernMax          int     `json:"uvHealthConcernMax"`
				UvHealthConcernMin          int     `json:"uvHealthConcernMin"`
				UvIndexAvg                  int     `json:"uvIndexAvg"`
				UvIndexMax                  int     `json:"uvIndexMax"`
				UvIndexMin                  int     `json:"uvIndexMin"`
				VisibilityAvg               float64 `json:"visibilityAvg"`
				VisibilityMax               float64 `json:"visibilityMax"`
				VisibilityMin               float64 `json:"visibilityMin"`
				WeatherCodeMax              int     `json:"weatherCodeMax"`
				WeatherCodeMin              int     `json:"weatherCodeMin"`
				WindDirectionAvg            float64 `json:"windDirectionAvg"`
				WindGustAvg                 float64 `json:"windGustAvg"`
				WindGustMax                 float64 `json:"windGustMax"`
				WindGustMin                 float64 `json:"windGustMin"`
				WindSpeedAvg                float64 `json:"windSpeedAvg"`
				WindSpeedMax                float64 `json:"windSpeedMax"`
				WindSpeedMin                float64 `json:"windSpeedMin"`
			} `json:"values"`
		} `json:"daily"`
	} `json:"timelines"`
}

// NewTomorrowIo returns a new TomorrowIo handle
func NewTomorrowIo(apiKey string) TomorrowIo {
	return TomorrowIo{apiKey: apiKey}
}

// Forecast returns the forecast for a given location
func (f *TomorrowIo) Forecast(l Location) (weather Weather, err error) {
	dlog.Printf("getting forecast for %#v", l)

	forecast, err := f.GetForecast(l)
	if err != nil {
		return
	}

	current, err := f.GetCurrent(l)
	if err != nil {
		return
	}

	weather.URL = fmt.Sprintf("%s?lat=%f&lon=%f&units=%s", tiAPI, l.Latitude, l.Longitude, tiUnits)

	weather.Current.Summary = tiDescriptions[current.Data.Values.WeatherCode]
	weather.Current.Icon = tiIconNames[current.Data.Values.WeatherCode]
	weather.Current.Humidity = current.Data.Values.Humidity
	weather.Current.Temp = temperature(current.Data.Values.Temperature)
	weather.Current.ApparentTemp = temperature(current.Data.Values.TemperatureApparent)

	for _, d := range forecast.Timelines.Daily {
		highTemp := d.Values.TemperatureMax
		lowTemp := d.Values.TemperatureMin

		dlog.Printf(">>> processing daily forecast for %s", d.Time)

		f := dailyForecast{
			Date:     parseTime(d.Time),
			Icon:     tiIconNames[d.Values.WeatherCodeMax],
			Summary:  tiDescriptions[d.Values.WeatherCodeMax],
			HighTemp: temperature(highTemp),
			LowTemp:  temperature(lowTemp),
			Sunrise:  parseTime(d.Values.SunriseTime),
			Sunset:   parseTime(d.Values.SunsetTime),
			Precip:   int(math.Round(d.Values.PrecipitationProbabilityAvg)),
		}
		weather.Daily = append(weather.Daily, f)

		dlog.Printf("initialized precip to %d\n", f.Precip)
	}

	for _, d := range forecast.Timelines.Hourly {
		f := hourlyForecast{
			Time:         parseTime(d.Time),
			Icon:         tiIconNames[d.Values.WeatherCode],
			Summary:      tiDescriptions[d.Values.WeatherCode],
			Temp:         temperature(d.Values.Temperature),
			ApparentTemp: temperature(d.Values.TemperatureApparent),
			Precip:       int(math.Round((d.Values.PrecipitationProbability))),
		}
		weather.Hourly = append(weather.Hourly, f)
	}

	return
}

func (f *TomorrowIo) GetForecast(l Location) (data tiForecast, err error) {
	dlog.Printf("getting hourly forecast for %#v", l)

	query := url.Values{}
	query.Set("location", fmt.Sprintf("%f,%f", l.Latitude, l.Longitude))
	query.Set("apikey", f.apiKey)
	query.Set("units", tiUnits)

	url := fmt.Sprintf("%s/weather/forecast?%s", tiAPI, query.Encode())

	var request *http.Request
	if request, err = http.NewRequest("GET", url, nil); err != nil {
		return
	}

	var resp *http.Response
	if resp, err = client.Do(request); err != nil {
		return
	}
	defer resp.Body.Close()

	if resp.StatusCode >= 400 {
		return data, fmt.Errorf(resp.Status)
	}

	err = json.NewDecoder(resp.Body).Decode(&data)
	return
}

func (f *TomorrowIo) GetCurrent(l Location) (data tiCurrent, err error) {
	dlog.Printf("getting current conditions for %#v", l)

	query := url.Values{}
	query.Set("location", fmt.Sprintf("%f,%f", l.Latitude, l.Longitude))
	query.Set("apikey", f.apiKey)
	query.Set("units", tiUnits)

	url := fmt.Sprintf("%s/weather/realtime?%s", tiAPI, query.Encode())

	dlog.Printf("getting URL %s", url)

	var request *http.Request
	if request, err = http.NewRequest("GET", url, nil); err != nil {
		return
	}

	var resp *http.Response
	if resp, err = client.Do(request); err != nil {
		return
	}
	defer resp.Body.Close()

	if resp.StatusCode >= 400 {
		return data, fmt.Errorf(resp.Status)
	}

	err = json.NewDecoder(resp.Body).Decode(&data)
	return
}

func parseTime(timeStr string) time.Time {
	loc := time.Now().Location()
	date, _ := time.Parse(time.RFC3339, timeStr)
	return date.In(loc)
}
